import re
from typing import Dict, Optional

from jira.resources import Sprint

from ..plugin import BaseFunction


class Function(BaseFunction):
    """ Returns a `jira.resources.Sprint` object for a given sprint ID string."""

    CACHE: Dict[int, str] = {}

    def get_sprint_details(self, representation: str) -> Sprint:
        # com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[id=436,rapidViewId=153,state=CLOSED,name=ORC #11 3/9-3/23,goal=Code Complete 3/18,startDate=2020-03-09T21:53:07.264Z,endDate=2020-03-23T20:53:00.000Z,completeDate=2020-03-23T21:08:29.391Z,sequence=436
        sprint_field_matcher = re.compile(
            r"com.atlassian.greenhopper.service.sprint.Sprint@.*id=(?P<id>\d+)[^\d].*"  # noqa
        )
        match = sprint_field_matcher.match(representation)
        if not match:
            return None

        id = int(match.groupdict()["id"])

        if id not in self.CACHE:
            self.CACHE[id] = self.jira.sprint(id)

        return self.CACHE[id]

    def __call__(self, sprint_id: Optional[str]) -> Optional[str]:  # type: ignore[override]
        if sprint_id is None:
            return None

        return self.get_sprint_details(sprint_id)



import argparse

from rich.console import Console
from rich.traceback import install as enable_rich_traceback

from .exceptions import UserError
from .plugin import get_installed_commands
from .utils import get_config, get_default_config_path


def main():
    enable_rich_traceback()
    commands = get_installed_commands()

    default_config_path = get_default_config_path()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        "-c",
        default=default_config_path,
        type=str,
        help=(
            "Path to configuration file to load; by default configuration "
            f"will be loaded from {default_config_path}."
        ),
    )
    parser.add_argument(
        "--instance-name",
        "-n",
        default="default",
        help=(
            "Named instance to connect to.  You can define new instances by "
            'running `jira-select configure --instance-name="my-instance"` '
            "to set the instance URL, username, and password for later use in "
            'other commands via providing the `--instance-name="my-instance" '
            "command-line argument.  By default the 'default' instance name "
            "will be used."
        ),
    )
    parser.add_argument(
        "--instance-url",
        "-i",
        default=None,
        type=str,
        help="URL of the Jira instance to connect to (if you cannot use 'configure')",
    )
    parser.add_argument(
        "--username",
        "-u",
        default=None,
        type=str,
        help="Username to use for connecting to Jira (if you cannot use 'configure')",
    )
    parser.add_argument(
        "--password",
        "-p",
        default=None,
        type=str,
        help=(
            "Password to use for connecting to Jira.  NOT RECOMMENDED: use the "
            "'store-password' or 'configure' command instead."
        ),
    )
    parser.add_argument(
        "--debugger",
        action="store_true",
        help=(
            "Wait for debugger connection before processing command (requires debugpy)."
        ),
    )
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    for cmd_name, cmd_class in commands.items():
        parser_kwargs = {}

        cmd_help = cmd_class.get_help()
        if cmd_help:
            parser_kwargs["help"] = cmd_help

        subparser = subparsers.add_parser(cmd_name, **parser_kwargs)
        cmd_class.add_arguments(subparser)

    args = parser.parse_args()

    if args.debugger:
        import debugpy

        debugpy.listen(("0.0.0.0", 5678))

        # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()

    config = get_config(path=args.config)

    console = Console()

    command = commands[args.command](config=config, options=args)

    try:
        command.handle()
    except UserError as e:
        console.print(f"[red]{e}[/red]")



import csv
from typing import Any, Dict, List

from ..plugin import BaseFormatter


class Formatter(BaseFormatter):
    @classmethod
    def get_file_extension(cls) -> str:
        return "csv"

    def _generate_fieldnames(self) -> List[str]:
        fields = []

        for field in self.executor.query.select:
            fields.append(field["column"])

        return fields

    def open(self):
        super().open()
        self.out = csv.DictWriter(self.stream, fieldnames=self._generate_fieldnames())
        self.out.writeheader()

    def writerow(self, row: Dict[str, Any]):
        self.out.writerow(row)



from typing import Any, Optional, List

from ..plugin import BaseFunction
from ..query import Result


class Function(BaseFunction):
    """ Returns value in row for field having the specified human-readable name.

    Note that jira-select provides the row under the variable name `_`.

    E.g::

        select
            - field_by_name(_, "Story Points")
        from: issues

    """

    FIELDS: Optional[List] = None

    def get_fields(self):
        if self.FIELDS is None:
            self.FIELDS = self.jira.fields()

        return self.FIELDS

    def __call__(self, row: Result, name: str) -> Optional[Any]:  # type: ignore[override]
        fields = self.get_fields()

        for field in fields:
            if field["name"] == name:
                return getattr(row, field["key"])

        return None



from unittest.mock import Mock

from dotmap import DotMap

from jira_select.types import QueryDefinition
from jira_select.query import Executor

from .base import JiraSelectTestCase


class JiraList(list):
    pass


class TestQuery(JiraSelectTestCase):
    def setUp(self):
        super().setUp()

        self.JIRA_ISSUES = JiraList(
            [
                DotMap(
                    {
                        "key": "ALPHA-1",
                        "fields": {
                            "issuetype": "Issue",
                            "summary": "My Ticket",
                            "project": "ALPHA",
                            "story_points": 1,
                        },
                    }
                ),
                DotMap(
                    {
                        "key": "ALPHA-3",
                        "fields": {
                            "issuetype": "Bug",
                            "summary": "Another Ticket",
                            "project": "ALPHA",
                            "story_points": 10,
                        },
                    }
                ),
                DotMap(
                    {
                        "key": "ALPHA-2",
                        "fields": {
                            "issuetype": "Issue",
                            "summary": "My Ticket #2",
                            "project": "ALPHA",
                            "story_points": 1,
                        },
                    }
                ),
            ]
        )
        self.JIRA_ISSUES.total = len(self.JIRA_ISSUES)

        self.mock_jira = Mock(search_issues=Mock(return_value=self.JIRA_ISSUES))

    def test_simple(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
        }

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {"key": "ALPHA-1",},
            {"key": "ALPHA-3",},
            {"key": "ALPHA-2",},
        ]

        assert expected_results == actual_results

    def test_sort_by(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
            "sort_by": ["story_points desc", "key"],
        }

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {"key": "ALPHA-3",},
            {"key": "ALPHA-1",},
            {"key": "ALPHA-2",},
        ]

        assert expected_results == actual_results

    def test_group_by_aggregation(self):
        query: QueryDefinition = {
            "select": ["issuetype", "len(key)"],
            "from": "issues",
            "group_by": ["issuetype"],
            "sort_by": ["len(key)"],
        }

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {"issuetype": "Bug", "len(key)": 1,},
            {"issuetype": "Issue", "len(key)": 2,},
        ]

        assert expected_results == actual_results

    def test_cap(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
            "cap": 1,
        }

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        assert len(actual_results) == 1

    def test_simple_wprogress(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
        }

        query = Executor(self.mock_jira, query, True)

        actual_results = list(query)
        expected_results = [
            {"key": "ALPHA-1",},
            {"key": "ALPHA-3",},
            {"key": "ALPHA-2",},
        ]

        assert expected_results == actual_results

    def test_sort_by_wprogress(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
            "sort_by": ["story_points desc", "key"],
        }

        query = Executor(self.mock_jira, query, True)

        actual_results = list(query)
        expected_results = [
            {"key": "ALPHA-3",},
            {"key": "ALPHA-1",},
            {"key": "ALPHA-2",},
        ]

        assert expected_results == actual_results

    def test_group_by_aggregation_wprogress(self):
        query: QueryDefinition = {
            "select": ["issuetype", "len(key)"],
            "from": "issues",
            "group_by": ["issuetype"],
            "sort_by": ["len(key)"],
        }

        query = Executor(self.mock_jira, query, True)

        actual_results = list(query)
        expected_results = [
            {"issuetype": "Bug", "len(key)": 1,},
            {"issuetype": "Issue", "len(key)": 2,},
        ]

        assert expected_results == actual_results

    def test_cap_wprogress(self):
        query: QueryDefinition = {
            "select": ["key"],
            "from": "issues",
            "cap": 1,
        }

        query = Executor(self.mock_jira, query, True)

        actual_results = list(query)
        assert len(actual_results) == 1



from logging import getLogger
from typing import Tuple

import keyring
from jira import JIRA, JIRAError
from prompt_toolkit.shortcuts import input_dialog, yes_no_dialog

from ..constants import APP_NAME
from ..exceptions import UserError
from ..plugin import BaseCommand

logger = getLogger(__name__)


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Interactively allows you to configure jira-select to connect "
            "to your Jira instance."
        )

    def collect_credentials(self) -> Tuple[str, str, str]:
        instance_url = input_dialog(
            title="Instance URL",
            text="Please enter your Jira instance URL (e.g. 'https://mycompany.jira.com/'):",
        ).run()
        if not instance_url:
            raise UserError("Cancelled")
        username = input_dialog(
            title="Username",
            text=(
                "Please enter your Jira username: \n\n(for Jira Cloud instances, you may need to generate a new API token to use as a password at https://id.atlassian.com/manage-profile/security/api-tokens)"
            ),
        ).run()
        if not username:
            raise UserError("Cancelled")
        password = input_dialog(
            title="Password",
            text=f"Please enter the password for {username}: ",
            password=True,
        ).run()
        if not password:
            raise UserError("Cancelled")

        return instance_url, username, password

    def handle(self) -> None:
        instance_url = ""
        username = ""
        password = ""

        while True:
            instance_url, username, password = self.collect_credentials()

            try:
                JIRA(
                    instance_url,
                    options={"agile_rest_path": "agile",},
                    basic_auth=(username, password),
                    max_retries=0,
                )
                break
            except JIRAError:
                result = yes_no_dialog(
                    title="Error connecting to Jira",
                    text=(
                        "A connection to Jira could not be established using "
                        "the credentials you provided; try again?"
                    ),
                ).run()
                if not result:
                    raise UserError("Aborted; configuration not saved.")

        self.config.setdefault("instances", {})[self.options.instance_name] = {
            "url": instance_url,
            "username": username,
        }
        self.save_config()

        store_password = yes_no_dialog(
            title="Save password?",
            text="Would you like to save this password to your system keyring?",
        ).run()
        if store_password:
            keyring.set_password(APP_NAME, instance_url + username, password)



import argparse
import inspect

from rich.table import Table

from ..plugin import BaseCommand, get_installed_functions
from ..utils import evaluate_expression


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return "Lists all available functions, their signatures, and description."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "search_terms",
            nargs="*",
            type=str,
            help="Case-insensitive search term for limiting displayed results.",
        )
        parser.add_argument(
            "--having",
            help=(
                "A 'having' expression to use for limiting the displayed results. "
                "Searchable fields are 'name' and 'description'. "
                "E.g.--having=\"'average' in description.lower()\". "
            ),
        )

    def handle(self):
        table = Table(title="functions")
        table.add_column(header="name", style="green")
        table.add_column(header="description", style="white")

        functions = get_installed_functions(self.jira)
        for name, fn in functions.items():
            docstring = fn.__doc__

            try:
                signature = (
                    f"[bright_cyan]{name}{inspect.signature(fn)}[/bright_cyan]\n\n"
                )
            except Exception:
                signature = ""

            description = signature + docstring.strip() + "\n\n"

            row = {
                "name": name,
                "description": description,
            }
            if self.options.search_terms:
                matches = True
                for option in self.options.search_terms:
                    if option.lower() not in str(row).lower():
                        matches = False
                        break
                if not matches:
                    continue
            if self.options.having:
                if not evaluate_expression(self.options.having, row, functions):
                    continue

            table.add_row(row["name"], row["description"])

        self.console.print(table)
